  /*  actions.c

    File with device namespace actions routines

    Copyright (C) 2001  Mike Sullivan

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Mike Sullivan may be reached by email at  sullivam@us.ibm.com

*/

/*
    Changes.
    Initial version Mike Sullivan   3/27/2001

*/
#include "scsiname.h"
#include <linux/major.h>

void create_symlink_entry(config_entry *centry,sg_entry *dentry);
void create_device_entry(config_entry *centry,sg_entry *dentry);
int make_path_dir(char *path);

/*
 * action_config_entry: perform action specified by config entry
 *     centry - User configuration entry that matches this device node
 *     sge    - scsi entry of matching device
 *
 * returns: void
 */
void action_config_entry(config_entry *centry, sg_entry *sge) {

	if (strlen(centry->type.s.symlink)) {
		create_symlink_entry(centry,sge);
	} else {
		create_device_entry(centry, sge);
	}
	return;
}
	
/*
 * create_symlink_entry: Create a user named symlink to the target device node
 *     centry - User configuration entry that matches this device node
 *
 * returns: void
 */
void create_symlink_entry(config_entry *centry, sg_entry *sge) {
	int status;

	TRACE(LOG_INFO, "create_symlink_entry, link=%s target=%d%d\n",
	      centry->type.s.symlink,sge->major,sge->minor);
	/* target kdev should match kdev published by kernel */
	status=dev_find_device((int)makedev(sge->major, sge->minor), centry->type.s.devname);
	if (!status) {
		/* build the path portion of the fully qualified symlink name */
		if (make_path_dir(centry->type.s.symlink)) {
			/* remove old symlink (if it exists) and create new one */
			unlink(centry->type.s.symlink);
			if (!symlink(centry->type.s.devname,centry->type.s.symlink)) {
				TRACE (LOG_INFO, "Action Taken. Created symlink %s to %s\n",
				       centry->type.s.symlink,centry->type.s.devname);
				/* mark the symlink as created to we know what to remove on device removal */
				centry->type.s.symlink_created=1;
			} else { 
				if (errno == EEXIST) {
					TRACE (LOG_INFO, "Action Taken. ERROR symlink %s ALREADY EXISTS!\n",
					       centry->type.s.symlink);
					centry->type.s.symlink_created=1;
				} else {
					TRACE (LOG_INFO, "Action Taken. ERROR symlink %s to %s creation failed errno %x\n",
					       centry->type.s.symlink,centry->type.s.devname,errno);
					centry->type.s.symlink_created=0;
				}
			}
		}
	} else {
		TRACE (LOG_INFO, "Error. Symlink target can not be found kdev %x\n",
		       makedev(sge->major,sge->minor));
	}
	
	return;
}

/*
 * create_device_entry: Create a user named device entry for the targeted device
 *     centry - User configuration entry that matches this device node
 *
 * returns: void
 */
void create_device_entry(config_entry *centry, sg_entry *sge) {
	int uid=-1, gid=-1;
	int mode=-1;
	struct stat statbuf;
	int status;
	int recreate=0;
	
	/* Check for existance of alias name, if found save the attributes */
	status=lstat(centry->type.s.alias, &statbuf);
	if (!status) {
		if (S_ISLNK(statbuf.st_mode)) {
			status=stat(centry->type.s.alias, &statbuf);
			if (!status) {
				mode=statbuf.st_mode;
			} else {
				mode=-1;
			}
			recreate=1;
		} else {
			mode=statbuf.st_mode;
		}
		uid=statbuf.st_uid;
		gid=statbuf.st_gid;
		if ( statbuf.st_rdev != makedev(sge->major,sge->minor)) {
			recreate=1;
		}
	} else {
		recreate=1;
	}
	
	if (recreate) {
		status=unlink(centry->type.s.alias);
		if (mode == -1) {
			/* create mode */
			mode = (0660 | (strcmp(sge->type,"BLK") ? S_IFCHR : S_IFBLK));
			TRACE(LOG_INFO, "Mode set to 0x%x\n",mode);
		}
		make_path_dir(centry->type.s.alias);
		status=mknod(centry->type.s.alias, mode, makedev(sge->major, sge->minor));
		if (status == -1) {
			TRACE(LOG_INFO, "Unable to mknod %s errno %x\n",centry->type.s.alias,errno);
		} else {
			TRACE(LOG_INFO, "Successful recreation of %s\n",centry->type.s.alias);
			TRACE(LOG_INFO, "Issuing chmod 0x%x for %s\n",mode,centry->type.s.alias);
			status=chmod(centry->type.s.alias,mode);
			if (status == -1) {
				TRACE(LOG_INFO, "Unable to reset mode 0x%x for %s\n",mode,centry->type.s.alias);
			}
			if (uid != -1) {
				chown(centry->type.s.alias,uid,gid);
			}
		}
	} else {
		TRACE(LOG_INFO, "Successful. Compatible device entry already exists\n");
	}
	return;
}



/*
 * make_path_dir: create_symlink_entry helper routine to create a intermediate directories to support fully qualified symlink name 
 *
 * returns:
 *     TRUE - All intermediate directories created successfully
 *     FALSE - Creation of intermediate directory failed
 */
int make_path_dir(char *path)
{
	char ch;
	char *ptr1, *ptr2;
	char path_copy[STRING_LENGTH];

	strcpy (path_copy, path);
	ptr1 = path_copy;
	while (TRUE) {
		if ( ( ptr2 = strchr (ptr1 + 1, '/') ) == NULL ) break;
		ch = *ptr2;
		*ptr2 = '\0';
		if (access (path_copy, F_OK) == 0) {
			*ptr2 = ch;
			ptr1 = ptr2;
			continue;
		}
		if (mkdir (path_copy,S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) != 0) {
			TRACE (LOG_ERR, "Error %x making directory %s\n",errno,path_copy);
			*ptr2 = ch;
			return (FALSE);
		}
		*ptr2 = ch;
		ptr1 = ptr2;
	}
	return (TRUE);
} 



