  /*  config.c

    helper functions for reading and processing scsiname config file 
    (/etc/scsiname.conf)

    Copyright (C) 2001  Mike Sullivan

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Mike Sullivan may be reached by email at  sullivam@us.ibm.com

*/

/*
    Changes.

*/
#include "scsiname.h"
struct _config_entry *config_entry_head=0;
struct _config_entry *template_entry_head=0;
int config_use_sg_names=0;

/* config_process_line - Given a line from the user config file, parse it and
 *                         add the info to the config entry.
 *  arguments:
 *    line         line read from user config file (ie. /etc/scsiname.conf)
 *    centry       Config entry
 *
 *  returns:
 *    0            Line processed
 *    1            Error encountered, line ignored
 */
int config_process_line (char *line, config_entry *centry) {
	char *key;
	char *value=line;
	misc_config_entry *mentry;
	int err;
	char tmp[255];
	
	key=strsep(&value,"=");
	if (key) {
		
		if (strcasecmp(key,"symlink") == 0) {
			if ( (centry->entrytype == CONFIG_ENTRY_TYPE_SPECIFIC) || (!centry->entrytype) ) {
				strcpy(centry->type.s.symlink,value);
				if (!centry->entrytype) {
					centry->entrytype=CONFIG_ENTRY_TYPE_SPECIFIC;
					centry->next=config_entry_head;
					config_entry_head=centry;
				}
			}
			return 0;
		}

		if (strcasecmp(key,"alias") == 0) {
			if ( (centry->entrytype == CONFIG_ENTRY_TYPE_SPECIFIC) || (!centry->entrytype) ) {
				strcpy(centry->type.s.alias,value);
				if (!centry->entrytype) {
					centry->entrytype=CONFIG_ENTRY_TYPE_SPECIFIC;
					centry->next=config_entry_head;
					config_entry_head=centry;
				}
			}
			return 0;
		}
		
		if ( (strcasecmp(key, "template_symlink") == 0) ) {
			if ( (centry->entrytype == CONFIG_ENTRY_TYPE_TEMPLATE) || (!centry->entrytype) ) {
				strcpy(centry->type.t.template_symlink,value);

				if (!centry->entrytype) {
					centry->entrytype=CONFIG_ENTRY_TYPE_TEMPLATE;
					centry->next=template_entry_head;
					template_entry_head=centry;
				}
			}
			return 0;
		}
		
		mentry=malloc(sizeof(misc_config_entry));
		if (mentry) {
			int i;
			strcpy(mentry->keyname,key);
			strcpy(mentry->valuestr,value);
			if ( (err=regcomp(&mentry->preg, mentry->valuestr,0)) != 0 ) {
				regerror (err, &mentry->preg, tmp, 255);
				TRACE (LOG_ERR, "config::config_process_line, error compiling regexp: \"%s\"\t%s\n", mentry->valuestr,tmp);
				return 1;
			}

			/* determine and load comparison function for this keyword */
			for (i=0; i<MAX_KEY_FUNCTIONS; i++) {
				if (strcmp(mentry->keyname, keyfns[i].key) == 0) {
					mentry->matchfn=keyfns[i].matchfn;
					break;
				}
			}
			if (i == MAX_KEY_FUNCTIONS) {
				/* default comparison function is a string comparison */
				mentry->matchfn=string_compare;
			}
			mentry->next=centry->misc_head;
			centry->misc_head=mentry;
		} else {
			TRACE (LOG_INFO, "config::config_process_line, Unable to malloc variable keyword entry\n");
			return 1;
		}
	} else {
		TRACE (LOG_INFO, "config::config_process_line, = separator not FOUND\n");
		return 1;
	}
	return 0;
}

/* 
 * config_read_config - Read the user provided configuration file
 *  arguments:
 *    location     fully qualified filename of configuration file (ie. /etc/scsiname.conf)
 *
 *  returns:
 *    0            OK
 *    1            Otherwise
*/
static int config_read_config (char *location)
{
	int lineno=0;
	FILE *fp;
	char buf[STRING_LENGTH];
	int new_entry=TRUE;
	config_entry *centry=0;
	
	if ( ( fp = fopen (location, "r") ) == NULL ) {
		TRACE (LOG_ERR, "config:config_read_config, error opening file: %s",location);
		return 1;
	}
	
	while (fgets (buf, STRING_LENGTH, fp) != NULL) {
		char *line;
		
		buf[strlen (buf) - 1] = '\0';
		/*  Skip leading whitespace  */
		for (line=buf; isspace (*line); ++line);
		if (line[0] == '\0') { 
			new_entry=TRUE;
			continue;
		}
		if (line[0] == '#') {
			continue;
		}
		if (new_entry == TRUE) {
			centry=malloc(sizeof(config_entry));
			if (!centry) {
				TRACE (LOG_INFO, "config::config_read_config, Could not alloc config entry for line %x, %s\n",lineno,line);
				return 1;
			}
			memset(centry,0,sizeof(config_entry));
			new_entry=FALSE;
		}
		config_process_line (line, centry);
		lineno++;
	}
	fclose (fp);
	centry=config_entry_head;
	TRACE (LOG_INFO, "Logging User Configuration Info:\n");
	if (!centry)
		TRACE(LOG_INFO, "\tNone\n");
	while(centry) {
		config_print_entry(centry);
		centry=centry->next;
		TRACE (LOG_INFO, "\n");
	} 
	centry=template_entry_head;
	TRACE (LOG_INFO, "Logging User Template Info:\n");
	if (!centry)
		TRACE(LOG_INFO, "\tNone\n");
	while(centry) {
		config_print_entry(centry);
		centry=centry->next;
		TRACE (LOG_INFO, "\n");
	} 
	return 0;
}


/* 
 * config_print_entry - Display Config entry
 *  arguments:
 *    centry        Config entry to display     
 */
void config_print_entry(config_entry *centry) {
	misc_config_entry *mentry;
	TRACE (LOG_INFO, "  Config Entry %p contains:\n",centry);
	if (centry->entrytype == CONFIG_ENTRY_TYPE_SPECIFIC) {
		if (strlen(centry->type.s.alias)) {
			TRACE (LOG_INFO, "\talias=%s\n",centry->type.s.alias); 
		}
		if (strlen(centry->type.s.symlink)) {
			TRACE (LOG_INFO, "\tsymlink=%s\n",centry->type.s.symlink); 
		}
	} else {
		if (strlen(centry->type.t.template_symlink)) {
			TRACE (LOG_INFO, "\ttemplate_symlink=%s\n",centry->type.t.template_symlink); 
		}
	}
	
	mentry=centry->misc_head;
	while(mentry) {
		TRACE (LOG_INFO, "\t%s=%s fn %p\n",mentry->keyname,mentry->valuestr,mentry->matchfn);
		mentry=mentry->next;
	}
}

int config_read_info() {
	struct stat statbuf;

	/* Read /etc/scsiname.conf file */
	if ( (!config_entry_head) && (!template_entry_head) ) { // Read config file first time in only 
		if (stat (CONFIG_FILE, &statbuf) != 0) {
			TRACE (LOG_INFO, "%s does not exist. sg name files will be used\n", CONFIG_FILE);
			config_use_sg_names=1;
			return 0;
		}
		if (statbuf.st_size == 0) {
			TRACE (LOG_INFO, "config::config_read_info helper exiting since %s is empty\n",CONFIG_FILE);
			return 1;
		}
		/* Read user configuration file */
		if (config_read_config(CONFIG_FILE) != 0) {
			TRACE (LOG_INFO, "config::config_read_info, unable to read config file %s\n",CONFIG_FILE);
			return 1;
		}
	}
	return 0;
}

/* 
 * config_create_default_entry - Create a default config entry using the sg name
 *  arguments:
 *    sge          sg entry containing default name
 *
 *  returns:
 *    0            ERROR
 *    centry       dummy config entry
*/
config_entry * config_create_default_entry(sg_entry *sge)
{
	config_entry *centry=0;
	
	centry=malloc(sizeof(config_entry));
	if (!centry) {
		TRACE (LOG_INFO, "config::config_create_default_entry, Could not alloc config entry\n");
		return 0;
	}
	memset(centry,0,sizeof(config_entry));
	centry->entrytype=CONFIG_ENTRY_TYPE_SPECIFIC;
	sprintf(centry->type.s.alias, "%s/%s",DEVROOT,sge->name);
	centry->next=config_entry_head;
	config_entry_head=centry;
	TRACE (LOG_INFO, "Logging User Configuration Info:\n");
	config_print_entry(centry);
	return centry;
}
